// Copyright (C) 2021  Nexedi SA and Contributors.
//                     Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package xbtree
// set of [lo,hi) Key ranges.

import (
	"fmt"
	"sort"
)

const traceRangeSet = false
const debugRangeSet = false

// KeyRange represents [lo,hi) Key range.
type KeyRange struct {
	lo  Key
	hi_ Key // NOTE _not_ hi) to avoid overflow at ∞;  hi = hi_ + 1
}

// RangedKeySet is set of Keys with adjacent keys coaleced into Ranges.
//
// Zero value represents empty set.
type RangedKeySet struct {
	// TODO rework to use BTree lo->hi_ instead if in practice in treediff,
	// and other usage places, N(ranges) turns out to be not small
	// (i.e. performance turns out to be not acceptable)
	rangev []KeyRange // lo↑
}


///* XXX temp - reenable

// Has returns whether key k belongs to the range.
func (r *KeyRange) Has(k Key) bool {
	return (r.lo <= k && k <= r.hi_)
}

// Add adds key k to the set.
func (S *RangedKeySet) Add(k Key) {
	S.AddRange(KeyRange{lo: k, hi_: k})
}

// Del removes key k from the set.
func (S *RangedKeySet) Del(k Key) {
	S.DelRange(KeyRange{lo: k, hi_: k})
}

// Has returns whether key k belongs to the set.
func (S *RangedKeySet) Has(k Key) bool {
	return S.HasRange(KeyRange{lo: k, hi_: k})
}

//*/


// AddRange adds range r to the set.
func (S *RangedKeySet) AddRange(r KeyRange) {
	if traceRangeSet {
		fmt.Printf("\n\nAddRange:\n")
		fmt.Printf("  S: %s\n", S)
		fmt.Printf("  r: %s\n", r)
		defer fmt.Printf("->u: %s\n", S)
	}

	S.verify()
	defer S.verify()

	// find first ilo: r.lo < [ilo].hi
	l := len(S.rangev)
	ilo := sort.Search(l, func(i int) bool {
		 return r.lo <= S.rangev[i].hi_
	})
	debugfRSet("\tilo: %d\n", ilo)

	if ilo == l { // not found
		S.rangev = append(S.rangev, r)
		l++
		debugfRSet("\tappend %s\t-> %s\n", r, S)
	}

	// find last jhi: [jhi].lo < r.hi
	jhi := ilo
	for ;; jhi++ {
		if jhi == l {
			break
		}
		if S.rangev[jhi].lo <= r.hi_ {
			continue
		}
		break
	}
	debugfRSet("\tjhi: %d\n", jhi)

	// entries in [ilo:jhi) ∈ [r.lo,r.hi) and should be merged into one
	if (jhi - ilo) > 1 {
		lo  := S.rangev[ilo].lo
		hi_ := S.rangev[jhi-1].hi_
		S.rangev = append(
			S.rangev[:ilo],		append([]KeyRange{
			KeyRange{lo, hi_}},
			S.rangev[jhi:]...)...)
		debugfRSet("\tmerge S[%d:%d]\t-> %s\n", ilo, jhi, S)
	}
	jhi = -1 // no longer valid

	// if [r.lo,r.hi) was outside of any entry - create new entry
	if r.hi_ < S.rangev[ilo].lo {
		S.rangev = append(
			S.rangev[:ilo],		append([]KeyRange{
			r},
			S.rangev[ilo:]...)...)
		debugfRSet("\tinsert %s\t-> %s\n", r, S)
	}

	// now we have covered entries merged as needed into [ilo]
	// extend this entry if r coverage is wider
	if r.lo < S.rangev[ilo].lo {
		S.rangev[ilo].lo = r.lo
		debugfRSet("\textend left\t-> %s\n", S)
	}
	if r.hi_ > S.rangev[ilo].hi_ {
		S.rangev[ilo].hi_ = r.hi_
		debugfRSet("\textend right\t-> %s\n", S)
	}

	// and check if we should merge it with right/left neighbours
	if ilo+1 < len(S.rangev) { // right
		if S.rangev[ilo].hi_+1 == S.rangev[ilo+1].lo {
			S.rangev = append(
				S.rangev[:ilo],		append([]KeyRange{
				KeyRange{S.rangev[ilo].lo, S.rangev[ilo+1].hi_}},
				S.rangev[ilo+2:]...)...)
			debugfRSet("\tmerge right\t-> %s\n", S)
		}
	}

	if ilo > 0 { // left
		if S.rangev[ilo-1].hi_+1 == S.rangev[ilo].lo {
			S.rangev = append(
				S.rangev[:ilo-1],	append([]KeyRange{
				KeyRange{S.rangev[ilo-1].lo, S.rangev[ilo].hi_}},
				S.rangev[ilo+1:]...)...)
			debugfRSet("\tmerge left\t-> %s\n", S)
		}
	}

	// done
}

// DelRange removes range r from the set.
func (S *RangedKeySet) DelRange(r KeyRange) {
	if traceRangeSet {
		fmt.Printf("\n\nDelRange:\n")
		fmt.Printf("  S: %s\n", S)
		fmt.Printf("  r: %s\n", r)
		defer fmt.Printf("->d: %s\n", S)
	}

	S.verify()
	defer S.verify()

	// find first ilo: r.lo < [ilo].hi
	l := len(S.rangev)
	ilo := sort.Search(l, func(i int) bool {
		 return r.lo <= S.rangev[i].hi_
	})
	debugfRSet("\tilo: %d\n", ilo)

	if ilo == l { // not found
		debugfRSet("\tnon-overlap right\n")
		return
	}

	// find last jhi: [jhi].lo < r.hi
	jhi := ilo
	for ;; jhi++ {
		if jhi == l {
			break
		}
		if S.rangev[jhi].lo <= r.hi_ {
			continue
		}
		break
	}
	debugfRSet("\tjhi: %d\n", jhi)

	if jhi == 0 {
		debugfRSet("\tnon-overlap left\n")
		return
	}

	// [ilo+1:jhi-1] should be deleted
	// [ilo] and [jhi-1] overlap with [r.lo,r.hi) - they shuold be deleted, or shrinked,
	// or split+shrinked if ilo==jhi-1 and r is inside [ilo]
	if jhi-ilo == 1 && S.rangev[ilo].lo < r.lo && r.hi_ < S.rangev[ilo].hi_ {
		x := S.rangev[ilo]
		S.rangev = append(
			S.rangev[:ilo],		append([]KeyRange{
			x, x},
			S.rangev[ilo+1:]...)...)
		jhi++
		debugfRSet("\tpresplit copy %s\t-> %s\n", x, S)
	}
	if S.rangev[ilo].lo < r.lo { // shrink left
		S.rangev = append(
			S.rangev[:ilo],		append([]KeyRange{
			KeyRange{S.rangev[ilo].lo, r.lo-1}},
			S.rangev[ilo+1:]...)...)
		ilo++
		debugfRSet("\tshrink [%d] left\t-> %s\n", ilo, S)
	}
	if r.hi_ < S.rangev[jhi-1].hi_ { // shrink right
		S.rangev = append(
			S.rangev[:jhi-1],	append([]KeyRange{
			KeyRange{r.hi_+1, S.rangev[jhi-1].hi_}},
			S.rangev[jhi:]...)...)
		jhi--
		debugfRSet("\tshrink [%d] right\t-> %s\n", jhi-1, S)
	}

	if (jhi - ilo) > 0 {
		S.rangev = append(
			S.rangev[:ilo],
			S.rangev[jhi:]...)
		debugfRSet("\tdelete S[%d:%d]\t-> %s\n", ilo, jhi, S)
	}

	// done
}

// HasRange returns whether all keys from range r belong to the set.
func (S *RangedKeySet) HasRange(r KeyRange) (yes bool) {
	if traceRangeSet {
		fmt.Printf("\n\nHasRange:\n")
		fmt.Printf("  S: %s\n", S)
		fmt.Printf("  r: %s\n", r)
		defer func() {
			fmt.Printf("->·: %v\n", yes)
		}()
	}

	S.verify()

	// find first ilo: r.lo < [ilo].hi
	l := len(S.rangev)
	ilo := sort.Search(l, func(i int) bool {
		 return r.lo <= S.rangev[i].hi_
	})
	debugfRSet("\tilo: %d\n", ilo)

	if ilo == l { // not found
		return false
	}

	// all keys from r are in S if r ∈ [ilo]
	return (S.rangev[ilo].lo <= r.lo && r.hi_ <= S.rangev[ilo].hi_)
}


// Union returns RangedKeySet(A.keys | B.keys).
func (A *RangedKeySet) Union(B *RangedKeySet) *RangedKeySet {
	U := A.Clone()
	U.UnionInplace(B)
	return U
}

// Difference returns RangedKeySet(A.keys \ B.keys).
func (A *RangedKeySet) Difference(B *RangedKeySet) *RangedKeySet {
	D := A.Clone()
	D.DifferenceInplace(B)
	return D
}

// XXX Intersection

func (A *RangedKeySet) UnionInplace(B *RangedKeySet) {
	A.verify()
	B.verify()
	defer A.verify()

	// XXX dumb
	for _, r := range B.rangev {
		A.AddRange(r)
	}
}

func (A *RangedKeySet) DifferenceInplace(B *RangedKeySet) {
	A.verify()
	B.verify()
	defer A.verify()

	// XXX dumb
	for _, r := range B.rangev {
		if len(A.rangev) == 0 {
			break
		}
		A.DelRange(r)
	}
}


// --------

// verify check RangedKeySet for internal consistency:
// - ranges must be not overlapping nor adjacent and ↑
func (S *RangedKeySet) verify() {
	// XXX !debug -> return ?

	var badv []string
	badf := func(format string, argv ...interface{}) {
		badv = append(badv, fmt.Sprintf(format, argv...))
	}
	defer func() {
		if badv != nil {
			emsg := fmt.Sprintf("S.verify: fail:\n\n")
			for _, bad := range badv {
				emsg += fmt.Sprintf("- %s\n", bad)
			}
			emsg += fmt.Sprintf("\nS: %s\n", S)
			panic(emsg)
		}
	}()

	hi_Prev := KeyMin
	for i, r := range S.rangev {
		hiPrev := hi_Prev + 1
		if i > 0 && !(hiPrev < r.lo) { // NOTE not ≤ - adjacent ranges must be merged
			badf("[%d]: !(hiPrev < r.lo)", i)
		}
		if !(r.lo <= r.hi_) {
			badf("[%d]: !(r.lo <= r.hi_)", i)
		}
		hi_Prev = r.hi_
	}
}

// Clone returns copy of the set.
func (orig *RangedKeySet) Clone() *RangedKeySet {
	klon := &RangedKeySet{}
	klon.rangev = append(klon.rangev, orig.rangev...)
	return klon
}

// Empty returns whether the set is empty.
func (S *RangedKeySet) Empty() bool {
	return len(S.rangev) == 0
}

// Equal returns whether A == B.
func (A *RangedKeySet) Equal(B *RangedKeySet) bool {
	if len(A.rangev) != len(B.rangev) {
		return false
	}
	for i, ra := range A.rangev {
		rb := B.rangev[i]
		if ra != rb {
			return false
		}
	}
	return true
}

// Clear removes all elements from the set.
func (S *RangedKeySet) Clear() {
	S.rangev = nil
}

// AllRanges returns slice of all key ranges in the set.
// XXX -> iter?
func (S *RangedKeySet) AllRanges() /*readonly*/[]KeyRange {
	return S.rangev
}

func (S RangedKeySet) String() string {
	s := "{"
	for i, r := range S.rangev {
		if i > 0 {
			s += " "
		}
		s += r.String()
	}
	s += "}"
	return s
}

func (r KeyRange) String() string {
        slo  := "-∞";  if r.lo  > KeyMin { slo = fmt.Sprintf("%v", r.lo) }
        shi  := "∞";   if r.hi_ < KeyMax { shi = fmt.Sprintf("%v", r.hi_+1) }
	return fmt.Sprintf("[%s,%s)", slo, shi)
}


func debugfRSet(format string, argv ...interface{}) {
	if !debugRangeSet {
		return
	}
	fmt.Printf(format, argv...)
}
