// Copyright (C) 2020-2021  Nexedi SA and Contributors.
//                          Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package xbtree
// tests for δbtail.go
// XXX doc  (2 ways of testing: explicit + allstructs), treegen py helper

import (
	"bufio"
	"context"
	"flag"
	"fmt"
	"io"
	"io/ioutil"
	"math/rand"
	"os"
	"os/exec"
	"reflect"
	"regexp"
	"sort"
	"strings"
	"testing"
	"time"

	"lab.nexedi.com/kirr/go123/exc"
	"lab.nexedi.com/kirr/go123/xerr"
	"lab.nexedi.com/kirr/neo/go/transaction"
	"lab.nexedi.com/kirr/neo/go/zodb"
	_ "lab.nexedi.com/kirr/neo/go/zodb/wks"

	"lab.nexedi.com/nexedi/wendelin.core/wcfs/internal/xzodb"
)

const kInf Key = 10000 // inf key (TeX hack)
//const kInf Key = KeyMax

// XXX move infrastructure -> δbtail_treegen_test.go ?

// TreeGenSrv represents connection to running `treegen ...` server.
type TreeGenSrv struct {
	argv      []string
	pysrv     *exec.Cmd      // spawned `treegen ...`
	pyin      io.WriteCloser // input to pysrv
	pyoutRaw  io.ReadCloser  // output from pysrv
	pyout     *bufio.Reader  // buffered ^^^
}

// TreeSrv represents connection to running `treegen trees` server.
//
// Create it with StartTreeSrv(zurl).
// - Commit(treeTopology) -> tid
type TreeSrv struct {
	*TreeGenSrv
	zurl	  string
	treeRoot  zodb.Oid       // oid of the tree treegen works on
	head      zodb.Tid       // last made commit
}

// AllStructsSrv represents connection to running `treegen allstructs` server.
//
// Create it with StartAllStructsSrv().
// - AllStructs(maxdepth, maxsplit, n, seed, kv1, kv2)
type AllStructsSrv struct {
	*TreeGenSrv
}

// StartTreeGenSrv spawns `treegen ...` server.
func StartTreeGenSrv(argv ...string) (_ *TreeGenSrv, hello string, err error) {
	defer xerr.Contextf(&err, "treesrv %v: start", argv)

	// spawn `treegen ...`
	tg := &TreeGenSrv{argv: argv}
	tg.pysrv = exec.Command("./testprog/treegen.py", argv...)
	tg.pyin, err = tg.pysrv.StdinPipe()
	if err != nil {
		return nil, "", err
	}
	tg.pyoutRaw, err = tg.pysrv.StdoutPipe()
	if err != nil {
		return nil, "", err
	}
	tg.pyout = bufio.NewReader(tg.pyoutRaw)
	tg.pysrv.Stderr = os.Stderr // no redirection

	err = tg.pysrv.Start()
	if err != nil {
		return nil, "", err
	}

	// wait for hello message and return it
	defer func() {
		if err != nil {
			tg.Close() // ignore error
		}
	}()
	defer xerr.Context(&err, "handshake")
	hello, err = tg.pyout.ReadString('\n')
	if err != nil {
		if err == io.EOF {
			err = io.ErrUnexpectedEOF
		}
		return nil, "", err
	}
	hello = strings.TrimSuffix(hello, "\n")

	return tg, hello, nil
}

// Close shutdowns treegen server.
func (tg *TreeGenSrv) Close() (err error) {
	defer xerr.Contextf(&err, "treegen %v: close", tg.argv)

	err1 := tg.pyin.Close()
	err2 := tg.pyoutRaw.Close()
	err3 := tg.pysrv.Wait()
	return xerr.Merge(err1, err2, err3)
}


// StartTreeSrv spawns `treegen trees` server.
func StartTreeSrv(zurl string) (_ *TreeSrv, err error) {
	defer xerr.Contextf(&err, "tree.srv %s: start", zurl)
	tgSrv, hello, err := StartTreeGenSrv("trees", zurl)
	if err != nil {
		return nil, err
	}

	tg := &TreeSrv{TreeGenSrv: tgSrv, zurl: zurl}
	defer func() {
		if err != nil {
			tgSrv.Close() // ignore error
		}
	}()

	// tree.srv start @<at> tree=<root>
	defer xerr.Contextf(&err, "invalid hello %q", hello)
	startRe := regexp.MustCompile(`^tree.srv start @([^ ]+) root=([^ ]+)$`)
	m := startRe.FindStringSubmatch(hello)
	if m == nil {
		return nil, fmt.Errorf("unexpected format")
	}
	tg.head, err = zodb.ParseTid(m[1]) // <at>
	if err != nil {
		return nil, fmt.Errorf("tid: %s", err)
	}
	tg.treeRoot, err = zodb.ParseOid(m[2]) // <root>
	if err != nil {
		return nil, fmt.Errorf("root: %s", err)
	}

	return tg, nil
}

// StartAllStructsSrv spawns `treegen allstructs` server.
func StartAllStructsSrv() (_ *AllStructsSrv, err error) {
	defer xerr.Context(&err, "allstructs.srv: start")

	tgSrv, hello, err := StartTreeGenSrv("allstructs")
	if err != nil {
		return nil, err
	}

	sg := &AllStructsSrv{TreeGenSrv: tgSrv}
	defer func() {
		if err != nil {
			tgSrv.Close() // ignore error
		}
	}()

	defer xerr.Contextf(&err, "invalid hello %q", hello)
	if hello != "# allstructs.srv start" {
		return nil, fmt.Errorf("unexpected format")
	}

	return sg, nil
}


// Commit creates new commit with underlying tree changed to specified tree topology.
func (tg *TreeSrv) Commit(tree string) (_ zodb.Tid, err error) {
	defer xerr.Contextf(&err, "tree.srv %s: commit %s", tg.zurl, tree)

	_, err = io.WriteString(tg.pyin, tree + "\n")
	if err != nil {
		return zodb.InvalidTid, err
	}

	reply, err := tg.pyout.ReadString('\n')
	if err != nil {
		if err == io.EOF {
			err = io.ErrUnexpectedEOF
		}
		return zodb.InvalidTid, err
	}
	reply = strings.TrimSuffix(reply, "\n")

	tid, err := zodb.ParseTid(reply)
	if err != nil {
		return zodb.InvalidTid, fmt.Errorf("invalid reply: %s", err)
	}
	tg.head = tid
	return tid, nil
}

// AllStructs returns response from `treegen allstructs`
func (tg *AllStructsSrv) AllStructs(kv1, kv2 map[Key]string, maxdepth, maxsplit, n int, seed int64) (_ []string, err error) {
	req := fmt.Sprintf("%d %d %d/%d %s %s ", maxdepth, maxsplit, n, seed, kvtxt(kv1), kvtxt(kv2))
	defer xerr.Contextf(&err, "allstructs.srv: %s", req)

	_, err = io.WriteString(tg.pyin, req + "\n")
	if err != nil {
		return nil, err
	}

	structv := []string{}
	for {
		reply, err := tg.pyout.ReadString('\n')
		if err != nil {
			if err == io.EOF {
				err = io.ErrUnexpectedEOF
			}
			return nil, err
		}
		reply = strings.TrimSuffix(reply, "\n")

		if reply == "# ----" {
			return structv, nil // end of response
		}
		if strings.HasPrefix(reply, "#") {
			continue // comment
		}

		structv = append(structv, reply)
	}
}


// RTree represents Tree node covering [lo, hi_] key range in its parent tree.
type RTree struct {
	oid    zodb.Oid
	parent *RTree
	// XXX +children?
}

// RBucket represents Bucket node covering [lo, hi_] key range in its Tree.
// NOTE it is not [lo,hi) but [lo,hi_] instead to avoid overflow at KeyMax.
type RBucket struct {
	oid     zodb.Oid
	parent  *RTree
	lo, hi_ Key	// XXX -> KeyRange ?
	kv      map[Key]string // bucket's k->v; values were ZBlk objects whose data is loaded instead.
}

// RBucketSet represents set of buckets covering whole [-∞,∞) range.
type RBucketSet []*RBucket // k↑


// Get returns RBucket which covers key k.
func (rbs RBucketSet) Get(k Key) *RBucket {
	i := sort.Search(len(rbs), func(i int) bool {
		return k <= rbs[i].hi_
	})
	if i == len(rbs) {
		panicf("BUG: key %v not covered;  coverage: %s", k, rbs.coverage())
	}

	rb := rbs[i]
	if !(rb.lo <= k && k <= rb.hi_) {
		panicf("BUG: get(%v) -> [%v, %v];  coverage: %s", k, rb.lo, rb.hi_, rbs.coverage())
	}

	return rb
}

// coverage returns string representation of rbs coverage structure.
func (rbs RBucketSet) coverage() string {
	if len(rbs) == 0 {
		return "ø"
	}
	s := ""
	for _, rb := range rbs {
		if s != "" {
			s += " "
		}
		s += fmt.Sprintf("[%v, %v]", rb.lo, rb.hi_)
	}
	return s
}

/* XXX kill
// holeIdx returns what should be ΔBtree.holeIdx for specified tracked key set.
func (rbs RBucketSet) holeIdx(tracked SetKey) SetKey {
	holes := SetKey{}
	for k := range tracked {
		_, keyin := rbs.Get(k).kv[k]
		if !keyin {
			holes.Add(k)
		}
	}
	return holes
}
*/

// trackSet returns what should be ΔBtree.trackSet for specified tracked key set.
func (rbs RBucketSet) trackSet(tracked SetKey) PPTreeSubSet {
	trackSet := PPTreeSubSet{}
	for k := range tracked {
		kb := rbs.Get(k)
		// trackSet explicitly records only regular buckets.
		// embedded buckets all have oid=zodb.InvalidOid and would lead to z
		newNode := false
		if kb.oid != zodb.InvalidOid {
			track, already := trackSet[kb.oid]
			if !already {
				track = &nodeInTree{parent: kb.parent.oid, nchild: 0}
				trackSet[kb.oid] = track
				newNode = true
			}
			if track.parent != kb.parent.oid {
				panicf("BUG: %s: B%s -> multiple parents: %s %s", rbs.coverage(), kb.oid, track.parent, kb.parent.oid)
			}
		}

		p := kb.parent
		for p != nil {
			ppoid := zodb.InvalidOid // oid of p.parent
			if p.parent != nil {
				ppoid = p.parent.oid
			}

			newParent := false
			pt, already := trackSet[p.oid]
			if !already {
				pt = &nodeInTree{parent: ppoid, nchild: 0}
				trackSet[p.oid] = pt
				newParent = true
			}
			if pt.parent != ppoid {
				panicf("BUG: %s: T%s -> multiple parents: %s %s", rbs.coverage(), p.oid, pt.parent, ppoid)
			}

			if newNode {
				pt.nchild++
			}
			newNode = newParent
			p = p.parent
		}
	}
	return trackSet
}



// XGetTree loads Tree from zurl@at->obj<root>.
// Tree values must be ZBlk whose data is returned instead of references to ZBlk objects.
// The tree is returned structured by buckets as
//
//	[] [lo,hi){k->v}  k↑
func XGetTree(db *zodb.DB, at zodb.Tid, root zodb.Oid) RBucketSet {
	defer exc.Contextf("%s: @%s: get tree %s", db.Storage().URL(), at, root)
	X := exc.Raiseif

	txn, ctx := transaction.New(context.Background())
	defer txn.Abort()
	zconn, err := db.Open(ctx, &zodb.ConnOptions{At: at}); X(err)

	xztree, err := zconn.Get(ctx, root); X(err)
	ztree, ok := xztree.(*Tree)
	if !ok {
		exc.Raisef("expected %s, got %s", xzodb.TypeOf(ztree), xzodb.TypeOf(xztree))
	}

	rbucketv := RBucketSet{}
	xwalkDFS(ctx, -kInf, kInf, ztree, func(rb *RBucket) {
		rbucketv = append(rbucketv, rb)
	})
	if len(rbucketv) == 0 { // empty tree -> [-∞, ∞){}
		etree   := &RTree{
				oid:    root,
				parent: nil,
		}
		ebucket := &RBucket{
				oid:    zodb.InvalidOid,
				parent: etree,
				lo:     -kInf,
				hi_:    kInf,
				kv:     map[Key]string{},
		}
		rbucketv = RBucketSet{ebucket}
	}
	return rbucketv
}

// xwalkDFS walks ztree in depth-first order emitting bvisit callback on visited bucket nodes.
func xwalkDFS(ctx context.Context, lo, hi_ Key, ztree *Tree, bvisit func(*RBucket)) {
	_xwalkDFS(ctx, lo, hi_, ztree, /*rparent*/nil, bvisit)
}
func _xwalkDFS(ctx context.Context, lo, hi_ Key, ztree *Tree, rparent *RTree, bvisit func(*RBucket)) {
	X := exc.Raiseif

	err := ztree.PActivate(ctx); X(err)
	defer ztree.PDeactivate()

	rtree := &RTree{oid: ztree.POid(), parent: rparent}

	// [i].Key ≤ [i].Child.*.Key < [i+1].Key  i ∈ [0, len([]))
	//
	// [0].Key       = -∞ ; always returned so
	// [len(ev)].Key = +∞ ; should be assumed so
	ev := ztree.Entryv()
	for i := range ev {
		xlo  := lo;  if i   > 0       { xlo  = ev[i].Key() }
		xhi_ := hi_; if i+1 < len(ev) { xhi_ = ev[i+1].Key() - 1 }

		tchild, ok := ev[i].Child().(*Tree)
		if ok {
			_xwalkDFS(ctx, xlo, xhi_, tchild, rtree, bvisit)
			continue
		}

		zbucket := ev[i].Child().(*Bucket)
		err = zbucket.PActivate(ctx); X(err)
		defer zbucket.PDeactivate()

		bkv := make(map[Key]string)
		bentryv := zbucket.Entryv()

		for _, __ := range bentryv {
			k  := __.Key()
			xv := __.Value()
			pv, ok := xv.(zodb.IPersistent)
			if !ok {
				exc.Raisef("[%d] -> %s;  want IPersistent", k, xzodb.TypeOf(xv))
			}
			data, err := ZGetBlkData(ctx, pv.PJar(), pv.POid())
			if err != nil {
				exc.Raisef("[%d]: %s", k, err)
			}
			bkv[k] = data
		}


		b := &RBucket{oid: zbucket.POid(), parent: rtree, lo: xlo, hi_: xhi_, kv: bkv}
		bvisit(b)
	}
}

// XGetδKV translates {k -> δ<oid>} to {k -> δ(ZBlk(oid).data)} according to db@at1..at2 snapshots.
func XGetδKV(db *zodb.DB, at1, at2 zodb.Tid, δkvOid map[Key]ΔValue) map[Key]Δstring {
	defer exc.Contextf("%s: get δkv %s..%s %v", db.Storage().URL(), at1, at2, δkvOid)
	X := exc.Raiseif

	txn, ctx := transaction.New(context.Background())
	defer txn.Abort()
	zconn1, err := db.Open(ctx, &zodb.ConnOptions{At: at1}); X(err)
	zconn2, err := db.Open(ctx, &zodb.ConnOptions{At: at2}); X(err)

	δkv := map[Key]Δstring{}
	for k, δvOid := range δkvOid {
		δkv[k] = Δstring{
			Old: xzgetBlkData(ctx, zconn1, δvOid.Old),
			New: xzgetBlkData(ctx, zconn2, δvOid.New),
		}
	}

	return δkv
}

// KAdjMatrix is adjacency matrix that describes how set of tracked keys
// changes (always grow) when tree topology is updated from A to B.
//
// Adjacency matrix
//
// A,  B  - topologies			ex T/B1    T2/B1-B3
// Av, Bv - topologies with values	ex T/B1:a  T2/B1:b-B3:c
//
// δ(Av, Bv) - full diff {k -> v} for changed keys; DEL = k -> ø
//             ex δ(T/B1:a, T2/B1:b-B3:c) = {-1:a +1:b +3:c}
//
// Δ(T, Av, Bv) - subset of δ(Av, Bv) corresponding to initial tracking set T
//             ex Δ({1}, T/B1:a, T2/B1:b-B3:c) = {-1:a +1:b}  (no +3:c)
//
// kadj(A,B) {} k -> {k'}:	- adjacency matrix
//	∃v1,v2:  k'∈ Δ({k}, Av1, Bv2)
//
//	ex kadj(T/B1, T2/B1-B3) = {1:{1} 3:{1,3} ∞:{1,3}}	k ∈ A+B+{∞}	XXX adjust to new kadj
//                                + {0:{1} 2:{1,3} + ... all possible keys}
//
// Δ(T, Av, Bv) = δ(Av, Bv)/kadj(A,B)[T]
//
// i.e. = δ(Av, Bv) for k: k ∈ U kadj(A,B)[·]
//                            ·∈T
//
// XXX adjacency matrix is symmetric:	XXX KAdj verifies this at runtime
//
//	kadj(A,B) == kadj(B,A)
type KAdjMatrix map[Key]SetKey

// Map returns kadj·keys .
func (kadj KAdjMatrix) Map(keys SetKey) SetKey {
	res := SetKey{}
	for k := range keys {
		to, ok := kadj[k]
		if !ok {
			panicf("kadj.Map: %d ∉ kadj\n\nkadj: %v", k, kadj)
		}
		res.Update(to)
	}
	return res
}

// KAdj computes adjacency matrix for t1 -> t2 transition.
//
// The set of keys for which kadj matrix is computed can be optionally provided.
// This set of keys defaults to allTestKeys(t1,t2).
//
// KAdj itself is verified by testΔBTail on entries with .kadjOK set.
func KAdj(t1, t2 *tTreeCommit, keysv ...SetKey) (kadj KAdjMatrix) {
	// assert KAdj(A,B) == KAdj(B,A)
	kadj12 := _KAdj(t1,t2, keysv...)
	kadj21 := _KAdj(t2,t1, keysv...)
	if !reflect.DeepEqual(kadj12, kadj21) {
		panicf("KAdj not symmetric:\nt1: %s\nt2: %s\nkadj12: %v\nkadj21: %v",
			t1.tree, t2.tree, kadj12, kadj21)
	}
	return kadj12
}

// XXX kill after holeIdx removal
func _KAdj_old(t1, t2 *tTreeCommit, keysv ...SetKey) (kadj KAdjMatrix) {
	var keys SetKey
	switch len(keysv) {
	case 0:
		keys = allTestKeys(t1, t2)
	case 1:
		keys = keysv[0]
	default:
		panic("multiple key sets on the call")
	}


	// kadj = {} k -> adjacent keys.
	// if k is tracked -> changes to adjacents must be in Update(t1->t2).
	kadj = KAdjMatrix{}
	for k := range keys {
		adj1 := SetKey{}
		adj2 := SetKey{}

		q1 := []Key{k}
		q2 := []Key{k}

		//tracef("\n")
		for len(q1) > 0 || len(q2) > 0 {
			//tracef("q1: %v\n", q1)
			//tracef("q2: %v\n", q2)
			if l1 := len(q1); l1 > 0 {
				k1 := q1[l1-1];  q1 = q1[:l1-1]
				if !adj1.Has(k1) {
					for k_ := range t1.xkv.Get(k1).kv {
						adj1.Add(k_)
						for k__ := range t2.xkv.Get(k_).kv {
							if !adj2.Has(k__) {
								q2 = append(q2, k__)
							}
						}
					}
				}
			}

			if l2 := len(q2); l2 > 0 {
				k2 := q2[l2-1];  q2 = q2[:l2-1]
				if !adj2.Has(k2) {
					for k_ := range t2.xkv.Get(k2).kv {
						adj2.Add(k_)
						for k__ := range t1.xkv.Get(k_).kv {
							if !adj1.Has(k__) {
								q1 = append(q1, k__)
							}
						}
					}
				}
			}
		}

		adj := SetKey{}; adj.Add(k); adj.Update(adj1); adj.Update(adj2)
		kadj[k] = adj
	}

	return kadj
}

func _KAdj(t1, t2 *tTreeCommit, keysv ...SetKey) (kadj KAdjMatrix) {
	var keys SetKey
	switch len(keysv) {
	case 0:
		keys = allTestKeys(t1, t2)
	case 1:
		keys = keysv[0]
	default:
		panic("multiple key sets on the call")
	}

	tracef("\n\n_KAdj\n")
	tracef("t1:   %s\n", t1.tree)
	tracef("t2:   %s\n", t2.tree)
	tracef("keys: %s\n", keys)

	// kadj = {} k -> adjacent keys.
	// if k is tracked -> changes to adjacents must be in Update(t1->t2).
	kadj = KAdjMatrix{}
	for k := range keys {
		adj1 := SetKey{}
		adj2 := SetKey{}

		q1 := &RangedKeySet{}; q1.Add(k)
		q2 := &RangedKeySet{}; q2.Add(k)
		done1 := &RangedKeySet{}
		done2 := &RangedKeySet{}

		tracef("\nk%d\n", k)
		for !q1.Empty() || !q2.Empty() {
			tracef("q1:   %s\tdone1:  %s\n", q1, done1)
			tracef("q2:   %s\tdone2:  %s\n", q2, done2)
			for _, r1 := range q1.AllRanges() {
				lo1 := r1.lo
				for {
					b1 := t1.xkv.Get(lo1)
					tracef("  b1: %s\n", b1)
					for k_ := range keys {
						if b1.lo <= k_ && k_ <= b1.hi_ {
							adj1.Add(k_)
						}
					}
					b1r := KeyRange{b1.lo, b1.hi_}
					done1.AddRange(b1r)
					// q2 |= (b1.keyrange \ done2)
					δq2 := &RangedKeySet{}
					δq2.AddRange(b1r)
					δq2.DifferenceInplace(done2)
					q2.UnionInplace(δq2)
					tracef("q2 += %s\t-> %s\n", δq2, q2)

					// continue with next right bucket until r1 coverage is complete
					lo1 = b1.hi_ + 1
					if r1.hi_ < lo1 {
						break
					}
				}
			}
			q1.Clear()

			for _, r2 := range q2.AllRanges() {
				lo2 := r2.lo
				for {
					b2 := t2.xkv.Get(lo2)
					tracef("  b2: %s\n", b2)
					for k_ := range keys {
						if b2.lo <= k_ && k_ <= b2.hi_ {
							adj2.Add(k_)
						}
					}
					b2r := KeyRange{b2.lo, b2.hi_}
					done2.AddRange(b2r)
					// q1 |= (b2.keyrange \ done1)
					δq1 := &RangedKeySet{}
					δq1.AddRange(b2r)
					δq1.DifferenceInplace(done1)
					q1.UnionInplace(δq1)
					tracef("q1 += %s\t-> %s\n", δq1, q1)

					// continue with next right bucket until r2 coverage is complete
					lo2 = b2.hi_ + 1
					if r2.hi_ < lo2 {
						break
					}
				}
			}
			q2.Clear()
		}

		adj := SetKey{}; adj.Add(k); adj.Update(adj1); adj.Update(adj2)	// XXX kill Add(k)
		kadj[k] = adj
	}

	return kadj
}


// xverifyΔBTail_Update verifies how ΔBTail handles ZODB update for a tree with changes in between t1->t2.
func xverifyΔBTail_Update(t *testing.T, subj string, db *zodb.DB, treeRoot zodb.Oid, t1, t2 *tTreeCommit) {
	// verify transition at1->at2 for all initial states of tracked {keys} from kv1 + kv2 + ∞

	t.Run(fmt.Sprintf("Update/%s→%s", t1.tree, t2.tree), func(t *testing.T) {
		allKeys := allTestKeys(t1, t2)
		allKeyv := allKeys.SortedElements()

		kadj12 := KAdj(t1, t2)

		// verify at1->at2 for all combination of initial tracked keys.
		for kidx := range IntSets(len(allKeyv)) {
			keys := SetKey{}
			for _, idx := range kidx {
				keys.Add(allKeyv[idx])
			}

// XXX allocates and keeps too much memory in -verylong
// XXX also not so useful as above "Update/t1->t2" ?
//			t.Run(fmt.Sprintf(" track=%s", keys), func(t *testing.T) {
				xverifyΔBTail_Update1(t, subj, db, treeRoot, t1.at,t2.at, t1.xkv,t2.xkv, t2.δZ, t2.δxkv, keys, kadj12)
//			})
		}
	})
}

// xverifyΔBTail_Update1 verifies how ΔBTail handles ZODB update at1->at2 from initial
// tracked state defined by initialTrackedKeys.
func xverifyΔBTail_Update1(t *testing.T, subj string, db *zodb.DB, treeRoot zodb.Oid, at1,at2 zodb.Tid, xkv1,xkv2 RBucketSet, δZ *zodb.EventCommit, d12 map[Key]Δstring, initialTrackedKeys SetKey, kadj KAdjMatrix) {
	X := exc.Raiseif
	tracef("\n>>> Track=%s\n", initialTrackedKeys)

	var kadjTracked SetKey = nil
	var δT, δTok map[Key]Δstring = nil, nil

	// badf queues error message to be reported on return.
	var badv []string
	badf := func(format string, argv ...interface{}) {
		badv = append(badv, fmt.Sprintf(format, argv...))
	}
	defer func() {
		if badv != nil {
			emsg := fmt.Sprintf("%s  ; tracked=%v :\n\n", subj, initialTrackedKeys)
			emsg += fmt.Sprintf("d12:  %v\nδTok: %v\nδT:   %v\n\n", d12, δTok, δT)
			emsg += fmt.Sprintf("kadj[Tracked]: %v\nkadj: %v\n\n", kadjTracked, kadj)
			emsg += strings.Join(badv, "\n")
			emsg += "\n"

			t.Fatal(emsg)
		}
	}()


	txn, ctx := transaction.New(context.Background())
	defer txn.Abort()

	// zconn, δbtail @at1 with initial tracked set
	zconn, err := db.Open(ctx, &zodb.ConnOptions{At: at1});	X(err)
	δbtail := NewΔBtail(zconn.At(), db)

	xtree, err := zgetNodeOrNil(ctx, zconn, treeRoot); X(err)
	var ztree *Tree // = nil if treeRoot was deleted
	if xtree != nil {
		ztree = xtree.(*Tree)
	}

	kadjTracked = SetKey{} // kadj[Tracked]  (all keys adjacent to tracked keys)
	for k := range initialTrackedKeys {
		if ztree != nil {
			_, ok, path, err := ZTreeGetBlkData(ctx, ztree, k);  X(err)
			err = δbtail.Track(k, ok, path); X(err)
		} else {
			// if treeRoot is deleted - add it to tracked set with every key
			// being a hole. This aligns with the following situation
			//
			//	T1 -> ø -> T2
			//
			// where after T1->ø, even though the tree becomes deleted, its root
			// continues to be tracked and all keys migrate to holes in the
			// tracking set. By aligning initial state to the same as after
			// T1->ø, we test what will happen on ø->T2.
			err = δbtail.track(k, false, []zodb.Oid{treeRoot});  X(err)
		}

		kadjTracked.Update(kadj[k])
	}

	// assert Tracked ∈ kadj[Tracked]  except ∞   XXX -> initialTrackedKeys.Difference(kadjTracked)
	trackedNotInKadj := SetKey{}
	for k := range initialTrackedKeys {
		if !kadjTracked.Has(k) {
			trackedNotInKadj.Add(k)
		}
	}
	trackedNotInKadj.Del(kInf)
	if len(trackedNotInKadj) > 0 {
		badf("BUG: Tracked ∉ kadj[Tracked]  ; extra=%v", trackedNotInKadj)
		return
	}

	//              k ∈ d12
	// k ∈ δT  <=>
	//              k ∈   U kadj[·]
	//                   ·∈tracking
	δTok = map[Key]Δstring{} // d12[all keys that should be present in δT]
	for k,δv := range d12 {
		if kadjTracked.Has(k) {
			δTok[k] = δv
		}
	}

/* XXX kill
	// verify δbtail.holeIdx  against @at1
	// holes1 = tracked1 \ kv1
	holes1 := xkv1.holeIdx(initialTrackedKeys)
	holeIdx := δbtail.holeIdxFor(treeRoot)
	if !reflect.DeepEqual(holes1, holeIdx.SetKey) {
		badf("δbtail.holeIdx1 wrong  ; holeIdx=%v  holeIdxOK=%v", holeIdx, holes1)
	}
*/

	ø := PPTreeSubSet{}

	// verify δbtail.trackSet against @at1
	// trackSet1 = xkv1[tracked1]
	trackSet1 := xkv1.trackSet(initialTrackedKeys)
//	if !reflect.DeepEqual(trackSet1, δbtail.trackSet) {
//		badf("δbtail.trackSet1 wrong:\n\thave: %v\n\twant: %v", δbtail.trackSet, trackSet1)
//	}
	if !δbtail.trackSet.Equal(ø) {
		badf("δbtail.trackSet1 wrong:\n\thave: %v\n\twant: %v", δbtail.trackSet, ø)
	}
	if !δbtail.trackNew.Equal(trackSet1) {
		badf("δbtail.trackNew1 wrong:\n\thave: %v\n\twant: %v", δbtail.trackNew, trackSet1)
	}
//	δbtail.assertTrack(t, "1", /*ø*/PPTreeSubSet{}, trackSet1)


	// δB <- δZ
	δB, err := δbtail.Update(δZ); X(err)
	if δB.Rev != δZ.Tid {
		badf("δB: rev != δZ.Tid  ; rev=%s  δZ.Tid=%s", δB.Rev, δZ.Tid)
		return
	}

/* XXX kill
	// verify δbtail.holeIdx  against @at2
	// holes2 = tracked2 \ kv2  ( = kadj[tracked1] \ kv2)
	holes2 := xkv2.holeIdx(kadjTracked)
	if !reflect.DeepEqual(holes2, holeIdx.SetKey) {
		badf("δbtail.holeIdx2 wrong  ; holeIdx=%v  holeIdxOK=%v", holeIdx, holes2)
	}
*/

	// verify δbtail.trackSet  against @at2
	// trackSet2 = xkv2[tracked2]  ( = xkv2[kadj[tracked1]]
	trackSet2 := xkv2.trackSet(kadjTracked)
//	if !reflect.DeepEqual(trackSet2, δbtail.trackSet) {
//		badf("δbtail.trackSet2 wrong:\n\thave: %v\n\twant: %v", δbtail.trackSet, trackSet2)
//	}
	if !δbtail.trackSet.Equal(trackSet2) {
		badf("δbtail.trackSet2 wrong:\n\thave: %v\n\twant: %v", δbtail.trackSet, trackSet2)
	}
	if !δbtail.trackNew.Equal(ø) {
		badf("δbtail.trackNew2 wrong:\n\thave: %v\n\twant: %v", δbtail.trackNew, ø)
	}
//	δbtail.assertTrack(t, "2", trackSet2, /*ø*/PPTreeSubSet{})


	// assert δB.ByRoot == {treeRoot -> ...}  if δTok != ø
	//                  == ø                  if δTok == ø
	rootsOK := SetOid{}
	if len(δTok) > 0 {
		rootsOK.Add(treeRoot)
	}
	roots := SetOid{}
	for root := range δB.ByRoot {
		roots.Add(root)
	}
	if !reflect.DeepEqual(roots, rootsOK) {
		badf("δB: roots != rootsOK  ; roots=%v  rootsOK=%v", roots, rootsOK)
	}
	_, inδB := δB.ByRoot[treeRoot]
	if !inδB {
		return
	}


	// δT <- δB
	δToid := δB.ByRoot[treeRoot]         // {} k -> δoid
	δT     = XGetδKV(db, at1,at2, δToid) // {} k -> δ(ZBlk(oid).data)

	// δT must be subset of d12.
	// changed keys, that are
	// - in tracked set	 -> must be present in δT
	// - outside tracked set -> may  be present in δT  (kadj gives exact answer)

	// δT is subset of d12
	for _, k := range sortedKeys(δT) {
		_, ind12 := d12[k]
		if !ind12 {
			badf("δT[%v] ∉  d12", k)
		}
	}

	// k ∈ tracked set ->  must be present in δT
	// k ∉ tracked set ->  may  be present in δT  (kadj gives exact answer)
	for _, k := range sortedKeys(d12) {
		_, inδT   := δT[k]
		_, inδTok := δTok[k]
		if inδT && !inδTok {
			badf("δT[%v] ∉  δTok", k)
		}

		if !inδT && inδTok {
			badf("δT    ∌  δTok[%v]", k)
		}

		if inδT {
			if δT[k] != d12[k] {
				badf("δT[%v] ≠  δTok[%v]", k, k)
			}
		}
	}
}

// assertTrack verifies that trackSet == trackSetOK.
// XXX place
func assertTrack(t *testing.T, subj string, trackSet, trackSetOK PPTreeSubSet) {
	t.Helper()
	eq := trackSet.Equal(trackSetOK)
	if !eq {
		t.Errorf("%s:\n\thave: %v\n\twant: %v", subj, trackSet, trackSetOK)
	}
}

// assertTrack verifies state of .trackSet and .trackNew.
// XXX place
func (δbtail *ΔBtail) assertTrack(t *testing.T, subj string, trackSetOK, trackNewOK PPTreeSubSet) {
	t.Helper()
	assertTrack(t, subj + ": trackSet", δbtail.trackSet, trackSetOK)
	assertTrack(t, subj + ": trackNew", δbtail.trackNew, trackNewOK)
}

// xverifyΔBTail_rebuild verifies δBtail.rebuild during t0->t1->t2 transition.
//
// t0->t1 exercises from-scratch rebuild,
// t1->t2 further exercises incremental rebuild.
func xverifyΔBTail_rebuild(t *testing.T, db *zodb.DB, treeRoot zodb.Oid, t0, t1, t2 *tTreeCommit) {
	// XXX handle DEL
	// XXX Update -> Track -> rebuild ...
	// XXX can start with non-existing tree

//	t1 := t2.prev
//	t0 := t1.prev
	t.Run(fmt.Sprintf("rebuild/%s→%s", t0.tree, t1.tree), func(t *testing.T) {
		t.Skip("TODO") // FIXME rebuild is currently broken

		tAllKeys := allTestKeys(t0, t1, t2)
		tAllKeyv := tAllKeys.SortedElements()

		// tid -> "at_i"
		xat := map[zodb.Tid]string{
			t0.at: "at0",
			t1.at: "at1",
			t2.at: "at2",
		}

		fmt.Printf("@%s:  %v\n", xat[t0.at], t0.xkv.Flatten())
		fmt.Printf("@%s:  %v\n", xat[t1.at], t1.xkv.Flatten())
		fmt.Printf("@%s:  %v\n", xat[t2.at], t2.xkv.Flatten())

		kadj01 := KAdj(t0,t1, allTestKeys(t0,t1,t2))
		kadj10 := KAdj(t1,t0, allTestKeys(t0,t1,t2))
		kadj21 := KAdj(t2,t1, allTestKeys(t0,t1,t2))
		kadj12 := KAdj(t1,t2, allTestKeys(t0,t1,t2))

		ø := PPTreeSubSet{}

		// verify t0 -> t1 Track(keys1) Rebuild -> t2 Track(keys2) Rebuild
		// for all combinations of keys1 and keys2
		for k1idx := range IntSets(len(tAllKeyv)) {
			keys1 := SetKey{}
			for _, idx1 := range k1idx {
				keys1.Add(tAllKeyv[idx1])
			}

			// δkv1_1 = t1.δxkv / kadj10(keys1)
			keys1_0 := kadj10.Map(keys1)
			δkv1_1 := map[Key]Δstring{}
			for k := range keys1_0 {
				δv, ok := t1.δxkv[k]
				if ok {
					δkv1_1[k] = δv
				}
			}

			t.Run(fmt.Sprintf(" T%s;R", keys1), func(t *testing.T) {
				δbtail := NewΔBtail(t0.at, db)

				// assert trackSet=ø, trackNew=ø, vδB=[]
				δbtail.assertTrack(t, "@at0", ø, ø)  // XXX + vδB

				xverifyΔBTail_rebuild_U(t, δbtail, t0, t1, xat, /*trackSet=*/ø)
				xverifyΔBTail_rebuild_TR(t, db, δbtail, t1, treeRoot, xat,
					// after Track(keys1)
					keys1,
					/*trackSet=*/ ø,
					/*trackNew=*/ t1.xkv.trackSet(keys1),

					// after rebuild
					/*trackSet=*/ t1.xkv.trackSet(keys1_0),
					/*vδT=*/ δkv1_1)

				t.Run((" →" + t2.tree), func(t *testing.T) {
					// tracked keys1 becomes tracked keys1_2 after Update(t1->t2)
//					keys1_2 := kadj12.Map(keys1)
					keys1_2 := kadj12.Map(kadj01.Map(keys1_0))
					xverifyΔBTail_rebuild_U(t, δbtail, t1, t2, xat, /*trackSet=*/t2.xkv.trackSet(keys1_2))

					// tRestKeys2 = tAllKeys - keys1
					tRestKeys2 := tAllKeys.Difference(keys1)
					tRestKeyv2 := tRestKeys2.SortedElements()
					for k2idx := range IntSets(len(tRestKeyv2)) {
						keys2 := SetKey{}
						for _, idx2 := range k2idx {
							keys2.Add(tRestKeyv2[idx2])
						}

						// find out full set of keys that should become tracked after
						// Track(keys2) + rebuild
						keys12R := keys1_2.Union(keys2)
						for {
							keys12R_ := kadj10.Map(kadj21.Map(keys12R))
							if keys12R.Equal(keys12R_) {
								break
							}
							keys12R = keys12R_
						}

						// δkv1_2 = t1.δxkv / kadj10(kadj21(kadj12(keys1) | keys2))
						// δkv2_2 = t2.δxkv / kadj10(kadj21(kadj12(keys1) | keys2))
//						keys12_2 := keys1_2.Union(keys2)
//						keys12_0 := kadj10.Map(kadj21.Map(keys12_2))
						fmt.Printf("\n\n\nKKK\nkeys1=%s  keys2=%s\n", keys1, keys2)
						fmt.Printf("keys12R=%s\n", keys12R)

						fmt.Printf("t0.xkv: %v\n", t0.xkv)
						fmt.Printf("t1.xkv: %v\n", t1.xkv)
						fmt.Printf("t2.xkv: %v\n", t2.xkv)
						fmt.Printf("kadj21: %v\n", kadj21)
						fmt.Printf("kadj12: %v\n", kadj12)
						fmt.Printf("t2.xkv.trackSet(%s) -> %s", keys12R, t2.xkv.trackSet(keys12R))
						fmt.Printf("\n\n\n")


						δkv1_2 := map[Key]Δstring{}
						δkv2_2 := map[Key]Δstring{}
						for k := range keys12R {
							δv1, ok := t1.δxkv[k]
							if ok {
								δkv1_2[k] = δv1
							}
							δv2, ok := t2.δxkv[k]
							if ok {
								δkv2_2[k] = δv2
							}
						}

						// t.Run is expensive at this level of nest
//						t.Run(" T"+keys2.String()+";R", func(t *testing.T) {
							δbtail_ := δbtail.clone()
							xverifyΔBTail_rebuild_TR(t, db, δbtail_, t2, treeRoot, xat,
								// after Track(keys2)
								keys2,
//								/*trackSet*/ t2.xkv.trackSet(keys1_0.Union(keys1_2)),
								/*trackSet*/ t2.xkv.trackSet(keys1_2),
								// FIXME vvv trackNew should not cover ranges that are already in trackSet
								/*trackNew*/ t2.xkv.trackSet(keys2),

								// after rebuild
								/* trackSet=*/ t2.xkv.trackSet(keys12R),
								/*vδT=*/ δkv1_2, δkv2_2)
//						})
					}
				})
			})
		}
	})
}

// xverifyΔBTail_rebuild_U verifies ΔBTree state after Update(ti->tj).
func xverifyΔBTail_rebuild_U(t *testing.T, δbtail *ΔBtail, ti, tj *tTreeCommit, xat map[zodb.Tid]string, trackSet PPTreeSubSet) {
	t.Helper()
	X := exc.Raiseif
	ø := PPTreeSubSet{}

	// Update ati -> atj
	δB, err := δbtail.Update(tj.δZ);  X(err)
	// XXX assert δB.Rev = tj.at; δB = δ(ti,tj)/initially tracked
	δbtail.assertTrack(t, fmt.Sprintf("after Update(@%s→@%s)", xat[ti.at], xat[tj.at]), trackSet, ø)
	_ = δB
}

// xverifyΔBTail_rebuild_TR verifies ΔBTree state after Track(keys) + rebuild.
func xverifyΔBTail_rebuild_TR(t *testing.T, db *zodb.DB, δbtail *ΔBtail, tj *tTreeCommit, treeRoot zodb.Oid, xat map[zodb.Tid]string, keys SetKey, trackSet PPTreeSubSet, trackNew, trackSetAfterRebuild PPTreeSubSet,vδTok ...map[Key]Δstring) {
	t.Helper()
	X := exc.Raiseif
	ø := PPTreeSubSet{}

	// Track(keys)
	txn, ctx := transaction.New(context.Background())
	defer txn.Abort()
	zconn, err := db.Open(ctx, &zodb.ConnOptions{At: tj.at});  X(err)
	xtree, err := zconn.Get(ctx, treeRoot);	 X(err)
	ztree := xtree.(*Tree)
	for k := range keys {
		_, ok, path, err := ZTreeGetBlkData(ctx, ztree, k);  X(err)
		err = δbtail.Track(k, ok, path);  X(err)
	}

	δbtail.assertTrack(t, fmt.Sprintf("@%s: after Track%v", xat[tj.at], keys), trackSet, trackNew)
	// XXX vδB=[ø]

	δbtail.rebuild()

	δbtail.assertTrack(t, fmt.Sprintf("@%s: after Track%v + rebuild", xat[tj.at], keys), trackSetAfterRebuild, ø)
	// XXX assert vδB=[δ1/T(keys)]

	// XXX verify Get

	// verify SliceByRootRev	XXX -> verify δbtail.byRoot[treeRoot] directly
	// XXX +KVAtTail, +lastRevOf
	l := len(vδTok)
	vatOK  := []zodb.Tid{}
	vδTok_ := []map[Key]Δstring{}
	t0 := tj
	for i := 0; i<l; i++ {
		// empty vδTok entries means they should be absent in vδT
		if δTok := vδTok[l-i-1]; len(δTok) != 0 {
			vatOK  = append([]zodb.Tid{t0.at}, vatOK...)
			vδTok_ = append([]map[Key]Δstring{δTok}, vδTok_...)
		}
		t0 = t0.prev
	}
	vδTok = vδTok_
	lo := t0.at
	hi := tj.at
	vδToid := δbtail.SliceByRootRev(treeRoot, lo, hi)

	l = len(vδToid)
	vat := make([]zodb.Tid, l)
	vδT := []map[Key]Δstring{}
	atPrev := lo
	for i, δToid := range vδToid {
		vat[i] = δToid.Rev
		δT := XGetδKV(db, atPrev, δToid.Rev, δToid.ΔKV) // {} k -> δ(ZBlk(oid).data)
		vδT = append(vδT, δT)
		atPrev = δToid.Rev
	}

	if !(reflect.DeepEqual(vat, vatOK) && reflect.DeepEqual(vδT, vδTok)) {
		have := ""
		for i := 0; i<len(vδT); i++ {
			have += fmt.Sprintf("\n\t@%s: %v", xat[vat[i]], vδT[i])
		}
		want := ""
		for i := 0; i<len(vδTok); i++ {
			want += fmt.Sprintf("\n\t@%s: %v", xat[vatOK[i]], vδTok[i])
		}
		t.Errorf("@%s: after Track%v + rebuild: SliceByRootRev:\nhave: %v\nwant: %v", xat[tj.at], keys, have, want)
	}
}

// xverifyΔBTail_GetAt verifies δBtail.Get on series of vt ZODB changes.
// XXX
// XXX kill
func ___xverifyΔBTail_GetAt(t *testing.T, db *zodb.DB, treeRoot zodb.Oid, vt ...*tTreeCommit) {
	subj := vt[0].tree
	for _, t := range vt[1:] {
		subj += "→" + t.tree
	}

	t.Run(fmt.Sprintf("Get/%s", subj), func(t *testing.T) {
		// tid -> "at_i"
		xat := map[zodb.Tid]string{}
		for i := range vt {
			xat[vt[i].at] = fmt.Sprintf("at%d", i)

			fmt.Printf("@%s:  %v\n", xat[vt[i].at], vt[i].xkv.Flatten())
		}

		tkeys := allTestKeys(vt...)
		tkeyv := tkeys.SortedElements()

		// verify  t1->t2-> ... ->tn Track(keys) Get(keys, @at)
		// for all combinations of tracked keys and at
		for kidx := range IntSets(len(tkeyv)) {
			keys := SetKey{}
			for _, idx := range kidx {
				keys.Add(tkeyv[idx])
			}

			t.Run(fmt.Sprintf("track=%s", keys), func(t *testing.T) {
				xverifyΔBTail_GetAt1(t, db, treeRoot, vt, xat, keys)
			})
		}
	})
}

func xverifyΔBTail_GetAt1(t *testing.T, db *zodb.DB, treeRoot zodb.Oid, vt []*tTreeCommit, xat map[zodb.Tid]string, keys SetKey) {
	X := exc.Raiseif

	// t1->t2-> ... -> tn
	δbtail := NewΔBtail(vt[0].at, db)
	for i := 1; i < len(vt); i++ {
		_, err := δbtail.Update(vt[i].δZ);  X(err)
	}

	// Track(keys)
	txn, ctx := transaction.New(context.Background())
	defer txn.Abort()
	zconn, err := db.Open(ctx, &zodb.ConnOptions{At: vt[len(vt)-1].at});  X(err)
	xtree, err := zconn.Get(ctx, treeRoot);	 X(err)
	ztree := xtree.(*Tree)

	for k := range keys {
		_, ok, path, err := ZTreeGetBlkData(ctx, ztree, k);  X(err)
		err = δbtail.Track(k, ok, path);  X(err)
	}

	// verify GetAt(k, @at) for all keys and @at
	for i := 1; i < len(vt); i++ {
		at := vt[i].at
		for _, k := range keys.SortedElements() {
			vOid, ok, rev, revExact, err := δbtail.GetAt(ctx, ztree, k, at);  X(err)
			v := xzgetBlkDataAt(db, vOid, rev)

			v_, ok_ := vt[i].xkv.Get(k).kv[k]
			rev_, revExact_ := vt[i].at, false
			for j := i-1; j >= 0; j-- {
				v__ := vt[j].xkv.Get(k).kv[k]
				if v__ != v_ {
					rev_ = vt[j+1].at
					revExact_ = true
					break
				}
				rev_ = vt[j].at
			}

			if v == "" { v = DEL }
			if v_ == "" { v_ = DEL }

			if !(v == v_ && ok == ok_ && rev == rev_ && revExact == revExact_) {
				t.Errorf("Get(%d, @%s) ->\nhave: %s, %v, @%s, %v\nwant: %s, %v, @%s, %v",
					k, xat[at],
					v, ok, xat[rev], revExact,
					v_, ok_, xat[rev_], revExact_)
			}
		}
	}
}


// ΔBTestEntry represents one entry in ΔBTail tests.
type ΔBTestEntry struct {
	tree   string     // next tree topology
	kadjOK KAdjMatrix // adjacency matrix against previous case (optional)
}

// ΔBTest converts xtest into ΔBTestEntry.
// xtest can be string|ΔBTestEntry.
func ΔBTest(xtest interface{}) ΔBTestEntry {
	var test ΔBTestEntry
	switch xtest := xtest.(type) {
	case string:
		test.tree   = xtest
		test.kadjOK = nil
	case ΔBTestEntry:
		test = xtest
	default:
		panicf("BUG: ΔBTest: bad type %T", xtest)
	}
	return test
}

// ΔBCommit represent test commit changing a tree.
type tTreeCommit struct {
	tree string            // the tree in toplogy-encoding
	prev *tTreeCommit      // previous commit
	at   zodb.Tid          // commit revision
	δZ   *zodb.EventCommit // raw ZODB changes; δZ.tid == at
	xkv  RBucketSet        // full tree state as of @at
	δxkv map[Key]Δstring   // full tree-diff against parent
}

// tZODBCacheEverything is workaround for ZODB/go not implementing real
// live cache for now: Objects get dropped on PDeactivate if cache
// control does not say we need the object to stay in the cache.
// XXX place
type tZODBCacheEverything struct{}
func (_ *tZODBCacheEverything) PCacheClassify(_ zodb.IPersistent) zodb.PCachePolicy {
	return zodb.PCachePinObject | zodb.PCacheKeepState
}

// testΔBTail verifies ΔBTail on sequence of tree topologies coming from testq.
func testΔBTail(t *testing.T, testq chan ΔBTestEntry) {
	X := exc.Raiseif

	work, err := ioutil.TempDir("", "δBTail"); X(err)
	defer func() {
		err := os.RemoveAll(work); X(err)
	}()
	tg, err := StartTreeSrv(work + "/1.fs"); X(err)
	defer func() {
		err := tg.Close(); X(err)
	}()

	zstor, err := zodb.Open(context.Background(), tg.zurl, &zodb.OpenOptions{
					ReadOnly: true,
		      }); X(err)
	defer func() {
		err := zstor.Close(); X(err)
	}()

	db := zodb.NewDB(zstor, &zodb.DBOptions{
		// We need objects to be cached, because otherwise it is too
		// slow to run the test for many testcases, especially
		// xverifyΔBTail_rebuild.
		CacheControl: &tZODBCacheEverything{},
	})
	defer func() {
		err := db.Close(); X(err)
	}()

	// XCommitTree calls tg.Commit and returns tTreeCommit corresponding to committed transaction.
	XCommitTree := func(tree string) *tTreeCommit {
		defer exc.Contextf("commit %s", tree)

		watchq := make(chan zodb.Event)
		at0 := zstor.AddWatch(watchq)
		defer zstor.DelWatch(watchq)

		tid, err := tg.Commit(tree); X(err)

		if !(tid > at0) {
			exc.Raisef("treegen -> %s  ; want > %s", tid, at0)
		}

		zevent := <-watchq
		δZ := zevent.(*zodb.EventCommit)
		if δZ.Tid != tid {
			exc.Raisef("treegen -> %s  ; watchq -> %s", tid, δZ)
		}

		// load tree structure from the db
		// if the tree does not exist yet - report its structure as empty
		var xkv RBucketSet
		if tree != DEL {
			xkv = XGetTree(db, δZ.Tid, tg.treeRoot)
		} else {
			// empty tree with real tg.treeRoot as oid even though the tree is
			// deleted.  Having real oid in the root tests that after deletion,
			// root of the tree stays in the tracking set. We need root to stay
			// in trackSet because e.g. in
			//
			//	T1 -> ø -> T2
			//
			// where the tree is first deleted, then recreated, without root
			// staying in trackSet after ->ø, treediff will notice nothing when
			// it comes to ->T2.
			xkv = RBucketSet{
				&RBucket{
					oid:    zodb.InvalidOid,
					parent: &RTree{
						oid:    tg.treeRoot, // NOTE oid is not InvalidOid
						parent: nil,
					},
					lo:     KeyMin,
					hi_:    KeyMax,
					kv:     map[Key]string{},
				},
			}
		}

		return &tTreeCommit{
			tree: tree,
			at:   δZ.Tid,
			δZ:   δZ,
			xkv:  xkv,
		}
	}

	var t0 *tTreeCommit
	t1 := &tTreeCommit{
		tree: "T/B:",	// treegen.py creates the tree as initially empty
		prev: nil,	// XXX ok?
		at:   tg.head,
		xkv:  XGetTree(db, tg.head, tg.treeRoot),
		δZ:   nil,	// XXX ok?
		δxkv: nil,	// XXX ok?
	}
	for test := range testq {
		t2 := XCommitTree(test.tree)
		t2.δxkv = kvdiff(t1.xkv.Flatten(), t2.xkv.Flatten())	// XXX move to XCommitTree?
		t2.prev = t1						// XXX ----//----

		subj := fmt.Sprintf("%s -> %s", t1.tree, t2.tree)
		tracef("\n\n\n**** %s ****\n\n", subj)

		// KAdj
		if kadjOK := test.kadjOK; kadjOK != nil {
			t.Run(fmt.Sprintf("KAdj/%s→%s", t1.tree, t2.tree), func(t *testing.T) {
				kadj := KAdj(t1, t2)
				if !reflect.DeepEqual(kadj, kadjOK) {
					t.Fatalf("BUG: computed kadj is wrong:\nkadjOK: %v\nkadj  : %v\n\n", kadjOK, kadj)
				}
			})
		}

		// ΔBTail.Update
		xverifyΔBTail_Update(t, subj, db, tg.treeRoot, t1,t2)

		// ΔBTail.rebuild
		if t0 != nil {
			xverifyΔBTail_rebuild(t, db, tg.treeRoot, t0,t1,t2)
		}

		t0, t1 = t1, t2
	}
}

// TestΔBTail verifies ΔBTail for explicitly provided tree topologies.
func TestΔBTail(t *testing.T) {
	// K is shorthand for SetKey
	K := func(keyv ...Key) SetKey {
		ks := SetKey{}
		for _, k := range keyv { ks.Add(k) }
		return ks
	}
	// oo is shorthand for kInf
	const oo = kInf
	// A is shorthand for KAdjMatrix
	type A = KAdjMatrix
	// Δ is shorthand for ΔBTestEntry
	Δ := func(tree string, kadjOK A) (test ΔBTestEntry) {
		test.tree   = tree
		test.kadjOK = kadjOK
		return test
	}

	// test known cases going through tree1 -> tree2 -> ...
	testv := []interface{} {
/*
		// start from non-empty tree to verify both ->empty and empty-> transitions
		"T/B1:a,2:b",

		// empty
		"T/B:",

		// +1
		Δ("T/B1:a",
			A{1:  K(1,oo),
			  oo: K(1,oo)}),

		// +2
		Δ("T/B1:a,2:b",
			A{1:  K(1,2,oo),
			  2:  K(1,2,oo),
			  oo: K(1,2,oo)}),

		// -1
		Δ("T/B2:b",
			A{1:  K(1,2,oo),
			  2:  K(1,2,oo),
			  oo: K(1,2,oo)}),

		// 2: b->c
		Δ("T/B2:c",
			A{2:  K(2,oo),
			  oo: K(2,oo)}),

		// +1 in new bucket (to the left)
		Δ("T2/B1:a-B2:c",
			A{1:  K(1,2,oo),
			  2:  K(1,2,oo),
			  oo: K(1,2,oo)}),

		// +3 in new bucket (to the right)
		Δ("T2,3/B1:a-B2:c-B3:c",
			A{1:  K(1),
			  2:  K(2,3,oo),
			  3:  K(2,3,oo),
			  oo: K(2,3,oo)}),

		// bucket split; +3 in new bucket
		"T/B1:a,2:b",
		Δ("T2/B1:a-B2:b,3:c",
			A{1:  K(1,2,3,oo),
			  2:  K(1,2,3,oo),
			  3:  K(1,2,3,oo),
			  oo: K(1,2,3,oo)}),

		// bucket split; +3 in new bucket; +4 +5 in another new bucket
		// everything becomes tracked because original bucket had [-∞,∞) coverage
		"T/B1:a,2:b",
		Δ("T2,4/B1:a-B2:b,3:c-B4:d,5:e",
			A{1:  K(1,2,3,4,5,oo),
			  2:  K(1,2,3,4,5,oo),
			  3:  K(1,2,3,4,5,oo),
			  4:  K(1,2,3,4,5,oo),
			  5:  K(1,2,3,4,5,oo),
			  oo: K(1,2,3,4,5,oo)}),
*/
		// reflow of keys: even if tracked={1}, changes to all B nodes need to be rescanned:
		// +B12 forces to look in -B23 which adds -3 into δ, which
		// forces to look into +B34 and so on.
		"T2,4,6/B1:a-B2:b,3:c-B4:d,5:e-B6:f,7:g",
		Δ("T3,5,7/B1:g,2:f-B3:e,4:d-B5:c,6:b-B7:a",
			A{1:  K(1,2,3,4,5,6,7),
			  2:  K(1,2,3,4,5,6,7),
			  3:  K(1,2,3,4,5,6,7),
			  4:  K(1,2,3,4,5,6,7),
			  5:  K(1,2,3,4,5,6,7),
			  6:  K(1,2,3,4,5,6,7),
			  7:  K(1,2,3,4,5,6,7),
			  oo: K(1,2,3,4,5,6,7,oo)}),

		// reflow of keys for rebuild: even if tracked1={}, tracked2={1}, changes to
		// all A/B/C nodes need to be rescanned. Contrary to the above case the reflow
		// is not detectable at separate diff(A,B) and diff(B,C) runs.
		"T3,5,7/B1:a,2:b-B3:c,4:d-B5:e,6:f-B7:g,8:h",
		"T/B1:b",
		"T2,4,6/B1:a-B2:b,3:c-B4:d,5:e-B6:f,7:g",
		// similar situation where rebuild has to detect reflow in between non-neighbour trees
		"T3,6/B1:a,2:b-B3:c,4:d-B6:f,7:g",
		"T4,7/B1:b-B4:d,5:e-B7:g,8:h",
		"T2,5,8/B1:a-B2:b,3:c-B5:e,6:f-B8:h,9:i",

		// depth=2; bucket split; +3 in new bucket; left T remain
		// _unchanged_ even though B under it is modified; right arm is
		// added to tracking set because 2 migrates there.
		"T/T/B1:a,2:b",
		Δ("T2/T-T/B1:a-B2:b,3:c",
			A{1:  K(1,2,3),
			  2:  K(1,2,3),
			  3:  K(1,2,3),
			  oo: K(1,2,3,oo)}),

		// depth=2; like prev. case, but additional right arm with +4
		// +5 is added which remain not tracked unless previously accessed.
		"T/T/B1:a,2:b",
		Δ("T2,4/T-T-T/B1:a-B2:b,3:c-B4:d,5:e",
			A{1:  K(1,2,3),
			  2:  K(1,2,3),
			  3:  K(1,2,3),
			  4:  K(1,2,3,4,5),
			  5:  K(1,2,3,4,5),
			  oo: K(1,2,3,4,5,oo)}),

		// depth=2; bucket split; +3 in new bucket; t0 and t1 split;
		// +right arm (T7/B45-B89) which remain not tracked unless previously accessed.
		"T/T/B1:a,2:b",
		Δ("T4/T2-T7/B1:a-B2:b,3:c-B4:d,5:e-B8:h,9:i",
			A{1:  K(1,2,3),
			  2:  K(1,2,3),
			  3:  K(1,2,3),
			  4:  K(1,2,3,4,5),
			  5:  K(1,2,3,4,5),
			  8:  K(1,2,3, 8,9),
			  9:  K(1,2,3, 8,9),
			  oo: K(1,2,3, 8,9,oo)}),


		// 2 reflow to right B neighbour; 8 split into new B; δ=ø
		"T3/B1:a,2:b-B4:d,8:h",
		"T2,5/B1:a-B2:b,4:d-B8:h", // XXX add A

		// tree deletion
		// having ø in the middle of the test cases exercises all:
		// * `ø -> Tree ...`		(tree is created anew),
		// * `... Tree -> ø`		(tree is deleted), and
		// * `Tree -> ø -> Tree`	(tree is deleted and then recreated)
		DEL,

		// found by AllStructs ([1] is not changed, but because B1 is
		// unlinked and 1 migrates to other bucket, changes in that
		// other bucket must be included into δT)
		"T1,2/B0:e-B1:d-B2:g,3:a",
		"T1/B0:d-B1:d,2:d",

		// XXX the same issue as ^^^ but with depth=2
		"T1,2/T-T-T/B0:a-B1:b-B2:c,3:d",
		"T1/T-T/B0:e-B1:b,2:f",


		// XXX depth=3 (to verify recursion and selecting which tree children to follow or not)

		// XXX more

		// degenerate topology from ZODB tests
		"T4/T2-T/T-T-T6,10/B1:a-B3:b-T-T-T/T-B7:c-B11:d/B5:e",
		"T/B1:e,5:d,7:c,8:b,11:a", // -3 +8

		// was leading treegen to generate corrupt trees
		"T/T1/T-T/B0:g-B1:e,2:d,3:h",
		"T1/T-T3/B0:g-T-T/B1:e,2:d-B3:h",

		// was leading to wrongly computed trackSet2 due to top not
		// being tracked to tree root.
		"T/T1/B0:a-B1:b",
		"T/T1/T-T/B0:c-B1:d",

		// was leading to wrongly computed trackSet2: leaf bucket not
		// reparented to root.
		"T/T/B0:a",
		"T/B0:a",


//		"T/T/T/B1:a,2:b",
//		"T/T/B1:a",

		// ---- found by AllStructs ----

		// trackSet2 wrongly computed due to top not being tracked to tree root
		"T2/T1-T/B0:g-B1:b-T/B2:b,3:a",
		"T2/T1-T/T-T-B2:a/B0:c-B1:g",

		// unchanged node is reparented
		"T1/B0:c-B1:f",
		"T1/T-T/B0:c-T/B1:h",

		// SIGSEGV in ApplyΔ
		"T1/T-T2/T-B1:c-B2:c/B0:g",
		"T1/T-T/B0:g-T/B1:e",

		// trackSet corruption: oid is pointed by some .parent but is not present
		"T1/T-T/B0:g-T2/B1:h-B2:g",
		"T/T1/T-T2/B0:e-B1:f-B2:g",

		// ApplyΔ -> xunion: node is reachable from multiple parents
		// ( because xdifference did not remove common non-leaf node
		//   under which there were also other changed, but not initially
		//   tracked, node )
		"T4/T1-T/T-T2-B4:c/T-T-T/B0:f-B1:h-B2:g,3:b",
		"T1/T-T/T-T2/T-T-T/B0:f-B1:h-B2:f",
		// ----//----
		"T3/T1-T/T-T2-T/B0:b-T-T-B3:h/B1:e-B2:a",
		"T1/T-T4/T-T2-T/T-T-T-T/B0:b-B1:e-B2:a,3:c-B4:e",
		// ----//----
		"T/T1,3/T-T2-T4/B0:b-T-T-B3:g-B4:c/B1:b-B2:e",
		"T1,4/T-T-T/T-T2-B4:f/T-T-T/B0:h-B1:b-B2:h,3:a",

		// XXX TODO classify
		"T2/B1:a-B7:g", "T2,8/B1:a-B7:g-B9:i",

		"T2/B1:a-B2:b", "T/B1:a,2:b",
		"T2,3/B1:a-B2:b-B3:c", "T/B1:a,2:b",
		"T2,3/B1:a-B2:c-B3:c", "T/B1:a,2:b",

		"T2/B1:a-B2:c", "T2,3/B1:a-B2:c-B3:c",

		"T2/B1:a-B3:c",
		Δ("T2/T-T4/B1:b-B3:d-B99:h",
			A{1:  K(1),
			  3:  K(3),
			  99: K(3,99),
			  oo: K(3,99,oo)}),
		// XXX --------

		// tree rotation
		"T3/B2:b-B3:c,4:d",
		"T5/T3-T7/B2:a-B3:a,4:a-B6:a-B8:a",
	}
	// direct  tree_i -> tree_{i+1} -> _{i+2} ...   plus
	// reverse ... tree_i <- _{i+1} <- _{i+2}
	kadjOK := ΔBTest(testv[len(testv)-1]).kadjOK
	for i := len(testv)-2; i >= 0; i-- {
		test := ΔBTest(testv[i])
		kadjOK, test.kadjOK = test.kadjOK, kadjOK
		testv = append(testv, test)
	}

	testq := make(chan ΔBTestEntry)
	go func() {
		defer close(testq)
		for _, test := range testv {
			testq <- ΔBTest(test)
		}
	}()
	testΔBTail(t, testq)

}

var (
	verylongFlag = flag.Bool("verylong", false, `switch tests to run in "very long" mode`)
	randseedFlag = flag.Int64("randseed", -1, `seed for random number generator`)
)

// TestΔBTailAllStructs verifies ΔBtail on tree topologies generated by AllStructs.
func TestΔBTailAllStructs(t *testing.T) {
	X := exc.Raiseif

	// considerations:
	// - depth↑  better for testing (more tricky topologies)
	// - nsplit↑ not so better for testing (leave s=1, max s=2)
	// - |kmin - kmax| affects N(variants) significantly
	//   -> keep key range small  (dumb increase does not help testing)
	// - N(keys) affects N(variants) significantly
	//   -> keep Nkeys reasonably small/medium  (dumb increase does not help testing)
	//
	// - spawning python subprocess is very slow (takes 300-500ms for
	//   imports; https://github.com/pypa/setuptools/issues/510)
	//   -> we spawn `treegen allstructs` once and use request/response approach.

	N := func(short, medium, long int) int {
		// -short
		if testing.Short() {
			return short
		}
		// -verylong
		if *verylongFlag {
			return long
		}
		// default
		return medium
	}
	maxdepth := N(2,  3,  4)
	maxsplit := N(1,  2,  2)
	n        := N(10,10,100)
	nkeys    := N(3,  5, 10)

	// server to generate AllStructs(kv1, kv2, ...)
	sg, err := StartAllStructsSrv(); X(err)
	defer func() {
		err := sg.Close(); X(err)
	}()

	// random seed
	seed := *randseedFlag
	if seed == -1 {
		seed = time.Now().UnixNano()
	}
	rng := rand.New(rand.NewSource(seed))
	t.Logf("# maxdepth=%d maxsplit=%d nkeys=%d n=%d seed=%d", maxdepth, maxsplit, nkeys, n, seed)

	// generate (kv1, kv2) pairs randomly

	// keysv1 and keysv2 are random shuffle of IntSets
	var keysv1 [][]int
	var keysv2 [][]int
	for keys := range IntSets(nkeys) {
		keysv1 = append(keysv1, keys)
		keysv2 = append(keysv2, keys)
	}
	v := keysv1
	rng.Shuffle(len(v), func(i,j int) { v[i], v[j] = v[j], v[i] })
	v = keysv2
	rng.Shuffle(len(v), func(i,j int) { v[i], v[j] = v[j], v[i] })

	// generate cases: keysv1[i] -> keysv2[i] with values generated
	// randomly along the way.
	vv := "abcdefgh"
	randv := func() string {
		i := rng.Intn(len(vv))
		return vv[i:i+1]
	}

	testq := make(chan ΔBTestEntry)
	go func() {
		defer close(testq)
		for i := range keysv1 {
			keys1 := keysv1[i]
			keys2 := keysv2[i]

			kv1 := map[Key]string{}
			kv2 := map[Key]string{}
			for k := range keys1 { kv1[Key(k)] = randv() }
			for k := range keys2 { kv2[Key(k)] = randv() }

			// given (kv1, kv2) - test on automatically generated (tree1 -> tree2)
			reqSeed := rng.Int63()
			treev, err := sg.AllStructs(kv1, kv2, maxdepth, maxsplit, n, reqSeed)
			if err != nil {
				t.Fatal(err)
			}

			for _, tree := range treev {
				testq <- ΔBTestEntry{tree, nil}
			}

		}
	}()

	testΔBTail(t, testq)
}



// ---- misc ----


// IntSets generates all sets of integers in range [0,N)
func IntSets(N int) chan []int {
	ch := make(chan []int)
	go intSets(ch, 0, N)
	return ch
}

// intSets generates all sets of integers in range [lo,hi)
func intSets(ch chan []int, lo, hi int) {
	ch <- nil // ø

	if lo < hi {
		for i := lo; i < hi; i++ {
			chTail := make(chan []int)
			go intSets(chTail, i+1, hi)
			for tail := range chTail {
				ch <- append([]int{i}, tail...) // i + tail
			}
		}
	}

	close(ch)
}

func TestIntSets(t *testing.T) {
	got := [][]int{}
	for is := range IntSets(3) {
		got = append(got, is)
	}

	I := func(v ...int) []int { return v }
	want := [][]int{I(),
			I(0), I(0,1), I(0,1,2), I(0,2),
			I(1), I(1,2),
			I(2),
	}

	if !reflect.DeepEqual(got, want) {
		t.Fatalf("error:\ngot:  %v\nwant: %v", got, want)
	}
}


// kvdiff returns difference in between kv1 and kv2.
var DEL = "ø" // DEL means deletion
type Δstring struct {
	Old string
	New string
}
func kvdiff(kv1, kv2 map[Key]string) map[Key]Δstring {
	delta := map[Key]Δstring{}
	keys := SetKey{}
	for k := range kv1 { keys.Add(k) }
	for k := range kv2 { keys.Add(k) }

	for k := range keys {
		v1, ok := kv1[k]
		if !ok { v1 = DEL }
		v2, ok := kv2[k]
		if !ok { v2 = DEL }
		if v1 != v2 {
			delta[k] = Δstring{v1,v2}
		}
	}

	return delta
}

func TestKVDiff(t *testing.T) {
	kv1 := map[Key]string{1:"a", 3:"c", 4:"d"}
	kv2 := map[Key]string{1:"b",        4:"d", 5:"e"}
	got  := kvdiff(kv1, kv2)
	want := map[Key]Δstring{1:{"a","b"}, 3:{"c",DEL}, 5:{DEL,"e"}}
	if !reflect.DeepEqual(got, want) {
		t.Fatalf("error:\ngot:  %v\nwant: %v", got, want)
	}
}

// kvtxt returns string representation of {} kv.
func kvtxt(kv map[Key]string) string {
	if len(kv) == 0 {
		return "ø"
	}

	keyv := []Key{}
	for k := range kv { keyv = append(keyv, k) }
	sort.Slice(keyv, func(i,j int) bool { return keyv[i] < keyv[j] })

	sv := []string{}
	for _, k := range keyv {
		v := kv[k]
		if strings.ContainsAny(v, " \n\t,:") {
			panicf("[%v]=%q: invalid value", k, v)
		}
		sv = append(sv, fmt.Sprintf("%v:%s", k, v))
	}

	return strings.Join(sv, ",")
}

func TestKVTxt(t *testing.T) {
	kv := map[Key]string{3:"hello", 1:"zzz", 4:"world"}
	got  := kvtxt(kv)
	want := "1:zzz,3:hello,4:world"
	if got != want {
		t.Fatalf("error:\ngot:  %q\nwant: %q", got, want)
	}
}

// Flatten converts xkv with bucket structure into regular dict.
func (xkv RBucketSet) Flatten() map[Key]string {
	kv := make(map[Key]string)
	for _, b := range xkv {
		for k,v := range b.kv {
			kv[k] = v
		}
	}
	return kv
}

// allTestKeys returns all keys from vt + ∞.
func allTestKeys(vt ...*tTreeCommit) SetKey {
	allKeys := SetKey{}; allKeys.Add(kInf) // ∞ simulating ZBigFile.Size() query
	for _, t := range vt {
		for _, b := range t.xkv {
			for k := range b.kv {
				allKeys.Add(k)
			}
		}
	}
	return allKeys
}

/*
// easies debugging / makes error output stable from run to run.
func (ks SetKey) SortedElements() []Key {
	keyv := ks.Elements()
	sort.Slice(keyv, func(i, j int) bool {
		return keyv[i] < keyv[j]
	})
	return keyv
}
func (ks SetKey) String() string {
	strv := []string{}
	for _, k := range ks.SortedElements() {
		strv = append(strv, fmt.Sprintf("%d", k))
	}
	return "{" + strings.Join(strv, " ") + "}"
}

func (os SetOid) SortedElements() []zodb.Oid {
	oidv := os.Elements()
	sort.Slice(oidv, func(i, j int) bool {
		return oidv[i] < oidv[j]
	})
	return oidv
}
func (os SetOid) String() string {
	strv := []string{}
	for _, oid := range os.SortedElements() {
		strv = append(strv, fmt.Sprintf("%s", oid))
	}
	return "{" + strings.Join(strv, " ") + "}"
}
*/

func sortedKeys(kv map[Key]Δstring) []Key {
	keyv := []Key{}
	for k := range kv {
		keyv = append(keyv, k)
	}
	sort.Slice(keyv, func(i, j int) bool {
		return keyv[i] < keyv[j]
	})
	return keyv
}

func (b *RBucket) String() string {
	// XXX dup wrt nodeInRange.String
	slo  := "-∞";  if b.lo  > KeyMin { slo = fmt.Sprintf("%v", b.lo) }
	shi  := "∞";   if b.hi_ < KeyMax { shi = fmt.Sprintf("%v", b.hi_+1) }
	return fmt.Sprintf("[%s,%s)B%s{%s}", slo, shi, b.oid, kvtxt(b.kv))

}

// String is like default %v, but uses ø for VDEL.
func (δv ΔValue) String() string {
	old, new := DEL, DEL
	if δv.Old != VDEL {
		old = δv.Old.String()
	}
	if δv.New != VDEL {
		new = δv.New.String()
	}
	return fmt.Sprintf("{%s %s}", old, new)
}


// ----------------------------------------

// ZBlk-related functions are imported at runtime by δbtail_x_test
var (
	ZTreeGetBlkData func(context.Context, *Tree, Key) (string, bool, []Node, error)
	ZGetBlkData     func(context.Context, *zodb.Connection, zodb.Oid) (string, error)
)

// xzgetBlkData loads block data from ZBlk object specified by its oid.
func xzgetBlkData(ctx context.Context, zconn *zodb.Connection, zblkOid zodb.Oid) string {
	X := exc.Raiseif

	if zblkOid == VDEL {
		return DEL
	}

	data, err := ZGetBlkData(ctx, zconn, zblkOid); X(err)
	return string(data)
}

// xzgetBlkDataAt loads block data from ZBlk object specified by oid@at.
func xzgetBlkDataAt(db *zodb.DB, zblkOid zodb.Oid, at zodb.Tid) string {
	X := exc.Raiseif

	txn, ctx := transaction.New(context.Background())
	defer txn.Abort()

	zconn, err := db.Open(ctx, &zodb.ConnOptions{At: at});  X(err)
	return xzgetBlkData(ctx, zconn, zblkOid)
}
